#pylint: disable=C0301, C0103, W0212, R0903, W0232

"""
.. module:: radical.pilot.url
   :platform: Unix
   :synopsis: Implementation of the Url class (forwarded from SAGA-Python).

.. moduleauthor:: Ole Weidner <ole.weidner@rutgers.edu>
"""

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

from saga import Url as sagaURL


# ------------------------------------------------------------------------------
#
class Url(sagaURL): 
    """Defines a URL class.
    """
    pass
